import sys
import logging
import urllib
import time
import subprocess
import json
import os

import splunk.util

import pdfgen_utils as pu

logger = pu.getLogger()

class Chart(object):
    _data = None
    _fields = None
    _props = None
    _svg = None
    _jsDirectory = None
    _pdfJsDirectory = None
    _runningAsScript = False

    EXPOSED_JS_DIRECTORY = "share/splunk/search_mrsparkle/exposed/js"
    PDF_JS_DIRECTORY = "share/splunk/pdf"

    def __init__(self, data, fields, props, runningAsScript=False):
        self._runningAsScript = runningAsScript
        self._data = data
        self._fields = fields
        self._props = props
        self._jsDirectory = self.getSplunkDirectory(self.EXPOSED_JS_DIRECTORY)
        self._pdfJsDirectory = self.getSplunkDirectory(self.PDF_JS_DIRECTORY)

    def getSplunkDirectory(self, relativePath):
        """ get a directory relative to the splunk home root """
        # TODO: this should be in splunk.util
        home = os.environ['SPLUNK_HOME']
        if not home or home == '/':
            raise Exception, 'readSplunkFile requires a SPLUNK_HOME to be set'

        workingPath = relativePath.strip(os.sep)
        workingPath = os.path.join(home, workingPath)
        pathParts = os.path.split(workingPath)
        pathParts = filter(lambda x: x != os.pardir, pathParts)
        finalPath = os.path.join(*pathParts)
        return finalPath

    def _buildArgs(self):
        args = {}
        args['props'] = self._props
        args['series'] = {}
        args['series']['fields'] = self._fields
        args['series']['columns'] = self._data
        self._log("_buildArgs, args: " + str(args))
        return json.dumps(args)

    def build(self):
        genSvgPath = os.path.join(self._pdfJsDirectory, 'gensvg.js')
        if 'NODE_PATH' not in os.environ:
            os.environ['NODE_PATH'] = os.path.join(os.environ['SPLUNK_HOME'], 'lib', 'node_modules')

        process = subprocess.Popen(['node', genSvgPath], stdout=subprocess.PIPE, stdin=subprocess.PIPE, stderr = subprocess.PIPE)
        argsJSON = self._buildArgs()
        logger.debug("argsJSON: %s" % argsJSON)

        out, err = process.communicate(input=argsJSON)

        self._svg = out
        if len(err) > 0:
            self._log("charting error logging: " + err, logLevel='warning')
        if len(out) == 0:
            self._log("no chart SVG output", logLevel='error')
        return len(out) > 0

    def getSvg(self):
        return self._svg

    def _log(self, msg, logLevel='debug'):
        if self._runningAsScript:
            print logLevel + " : " + msg
            return

        if logLevel=='debug':
            logger.debug(msg)
        elif logLevel=='info':
            logger.info(msg)
        elif logLevel=='warning':
            logger.warning(msg)
        elif logLevel=='error':
            logger.error(msg)


if __name__=="__main__":
    data = [[1, 2, 3, 4, 5], [-5, -4, -3, -2, -1]]
    props = {"chart": "line"}

    chart = Chart(data, props, runningAsScript=True)
    chart.build()
    print chart.getSvg()
